  $ cork-test help
  Usage: cork-test <command> [<options>]
  
  Available commands:
    c1       Command 1 (now with subcommands)
    c2       Command 2
    pwd      Print working directory
    mkdir    Create a directory
    rm       Remove a file or directory
    find     Search for a file in a list of directories
    paths    Print out standard paths for the current user
    dir      Print the contents of a directory
    sub      Run a subcommand
    cleanup  Test process cleanup functions
