/* -*- coding: utf-8 -*-
 * ----------------------------------------------------------------------
 * Copyright © 2011-2013, RedJack, LLC.
 * All rights reserved.
 *
 * Please see the COPYING file in this distribution for license details.
 * ----------------------------------------------------------------------
 */

#ifndef LIBCORK_CORE_H
#define LIBCORK_CORE_H

/*** include all of the parts ***/

#include <libcork/core/allocator.h>
#include <libcork/core/attributes.h>
#include <libcork/core/byte-order.h>
#include <libcork/core/callbacks.h>
#include <libcork/core/error.h>
#include <libcork/core/gc.h>
#include <libcork/core/hash.h>
#include <libcork/core/id.h>
#include <libcork/core/mempool.h>
#include <libcork/core/net-addresses.h>
#include <libcork/core/timestamp.h>
#include <libcork/core/types.h>
#include <libcork/core/u128.h>

#endif /* LIBCORK_CORE_H */
