/* -*- coding: utf-8 -*-
 * ----------------------------------------------------------------------
 * Copyright © 2011, RedJack, LLC.
 * All rights reserved.
 *
 * Please see the COPYING file in this distribution for license
 * details.
 * ----------------------------------------------------------------------
 */

#ifndef LIBCORK_DS_H
#define LIBCORK_DS_H

/*** include all of the parts ***/

#include <libcork/ds/array.h>
#include <libcork/ds/bitset.h>
#include <libcork/ds/buffer.h>
#include <libcork/ds/dllist.h>
#include <libcork/ds/hash-table.h>
#include <libcork/ds/managed-buffer.h>
#include <libcork/ds/ring-buffer.h>
#include <libcork/ds/slice.h>
#include <libcork/ds/stream.h>

#endif /* LIBCORK_DS_H */
