/* -*- coding: utf-8 -*-
 * ----------------------------------------------------------------------
 * Copyright © 2011, RedJack, LLC.
 * All rights reserved.
 *
 * Please see the COPYING file in this distribution for license
 * details.
 * ----------------------------------------------------------------------
 */

#define CORK_HASH_ATTRIBUTES  CORK_API

#include "libcork/core/hash.h"
#include "libcork/core/types.h"

/* All of the following functions will be defined for us by libcork/core/hash.h:
 *   cork_hash_buffer
 *   cork_big_hash_buffer
 *   cork_stable_hash_buffer
 */
